function [result,overflow] = ALU(op1,op2,fxd, opcode,ovMode, DBG,dbgStr)
%ALU	simulate the behavior of the VHDL description of an ALU
%	result = ALU(op1,op2,fxd, opcode) returns the result of the operation
%	as specified in 'opcode'. In here, 'op1' and 'op2' are two signed fixed- 
%	point input variables, 'result' is the output in the same signed fixed- 
%	point format.
%	This format has to be given in the vector 'fxd' as [N M], where N denotes
%	the total number of bits, while M defines the number of fractional bits.
%	Valid opcodes are 'add', 'sub', 'gt', 'ge', 'eq', 'le', 'lt' and 'nop'.
%
%	[result,overflow] = ALU(op1,op2,fxd, opcode,ovMode) also signals the 
%	correctness of the result in 'overflow'. A zero means that the result is 
%	unaltered and correct, overflow = 1 means that the result has been
%	'wrapped' (e.g. bits left of msb has been skipped) if ovMode = 'wrap' 
%	(the default choice), or has been saturated to its highest positive
%	value or lowest negative values when ovMode = 'sat' has been specified.  
%
%	See also MUL. 

% (c) H.J. Lincklaen Arriens, 
%     Delft University of Technology, November 2004


N = fxd(1);
M = fxd(2);
if ( N <= M )
	error( 'Incorrect values for parameter Fixed ...' );
end

maxPos  =  2^(N-1) -1;
maxNeg  = -2^(N-1);
fracFac =  2^M;

if ~exist('ovMode','var')
	ovMode = 'wrap';	% or 'sat'
end
if ~exist('dbgStr','var')
	dbgStr = '';
end

op1Int = op1 * fracFac;
if ( floor(op1Int) ~= op1Int )
    warning( 'ALU: op1 doesn''t fit in %d fractional bits ...', M );
	op1Int = floor(op1Int);
	warning( 'ALU: op1 truncated to %15.10f ...', op1Int/fracFac );
end 
if ((op1Int > maxPos) || (op1Int < maxNeg))
	error( 'ALU ERROR: signed op1 doesn''t fit in %d integer bits ...', N-M );
end

op2Int = op2 * fracFac;
if ( floor(op2Int) ~= op2Int )
	warning( 'ALU: op2 doesn''t fit in %d fractional bits ...', M );
	op2Int = floor(op2Int);
	warning( 'ALU: signed op2 truncated to %15.10f ...', op2Int/fracFac );
end 
if ((op2Int > maxPos) || (op2Int < maxNeg))
	error( 'ALU ERROR: signed op2 doesn''t fit in %d integer bits ...', N-M );
end

overflow = 0;
switch lower(opcode)
	case 'add'
		[result,overflow] = testWrap( op1Int + op2Int, maxPos,maxNeg,2^N, ovMode);
		result = floor(result)/fracFac;
		operStr = '+';
	case 'sub'
		[result,overflow] = testWrap( op1Int - op2Int, maxPos,maxNeg,2^N, ovMode);
		result = floor(result)/fracFac;
		operStr = '-';
	case 'gt'
		result = (op1Int > op2Int);
		operStr = '>';
	case 'ge'
		result = (op1Int >= op2Int);
		operStr = '>=';
	case 'eq'
		result = (op1Int == op2Int);
		operStr = '==';
	case 'le'
		result = (op1Int <= op2Int);
		operStr = '<=';
	case 'lt'
		result = (op1Int < op2Int);
		operStr = '<';
	otherwise
		result = 0;
		operStr = '?';
end
if overflow
	warning( 'overflow ...' );
end
if DBG
	fprintf( '%s   %s = %s %s %s\n', dbgStr, fixp2hex(result,fxd), ...
							fixp2hex(op1,fxd), operStr, fixp2hex(op2,fxd) ); 
end			

%================================================================================
%=======   LOCAL FUNCTIONS   ==================================================== 
%================================================================================

function [outVal,ovErr] = testWrap(testVal, maxPos,maxNeg,wrapVal, ovMode)
  ovErr = 0;
  if (testVal > maxPos)
    if strcmp(ovMode,'sat')
	  outVal = maxPos;
	else
	  outVal = testVal - wrapVal;
	end
	ovErr = 1;
  elseif (testVal < maxNeg)
    if strcmp(ovMode,'sat')
	  outVal = maxNeg;
	else
	  outVal = wrapVal + testVal;
	end
	ovErr = 1;
  else
    outVal = testVal;
  end
			