function coeffsSeq = gen_mTB(cirFilename,toScreen,schedMethod,varargin)
%GEN_MTB	Create MATLAB reference testbench files
%	GEN_MTB(cirFilename,toScreen,schedMethod,varargin) creates two MATLAB
%	files. If 'toScreen' is 1, these are written to screen; if 0, they
%	are written to disk. If 'cirFilename' is NAME.cir, they can be found in
%	(a newly created directory) NAME\matlab. The (top-level) testbench-file
%	will be named 'testbench_NAME_auto.m' (see its HELP function). 
%	Input data for the testbench is expected to be found (line by line) 
%	in a (user created) file named NAME.INP, and output will be written 
%	to a file NAME.OUT, both in the directory NAME\matlab. 
%	'schedMethod' specifies the scheduling method to be used, and can be
%	'ASAP', 'ALAP', FORCED' or 'LIST'.
%	The parameters that follow 'schedMethod' are dependant on the method
%	choosen, and may be 'delayMUL','delayALU','nMULs' and/or 'nALUs'.
%	-	'delayMUL' and 'delayALU' are the latencies of resp. MULtipliers and 
%		ALUs in integer multiples of a clock cycle. They are optional for 
%		'ASAP' and 'ALAP' (default, if not specified, is 1 cycle each). 
%		At this moment they are not needed for 'forceD' (both fixed to 1).
%		For the 'LIST' method, each of the delays need to be specified.
%	-	'nMULs' and 'nALUs' are the number of available MULtipliers and ALUs,
%		and have to be specified for the 'LIST' method only.
%
% 	coeffsSeq = GEN_MTB(cirFilename,toScreen,schedMethod,varargin) also
%	returns the names and the order of the coefficients to be expected in 
%	the .INP-file.
%
%	examples:	gen_mTB('my_file.cir',1,'asap')
%				gen_mTB('my_file.cir',0,'alap',2,1)
%				gen_mTB('my_file.cir',0,'forceD')
%				coeffsSeq = gen_mTB('my_file.cir',0,'list',1,1,3,2)
%
%	NOTE: 	GEN_MTB expects to be run from the directory in which the
%			.cir-file resides.
% 
%	See also GEN_VHD_TEMPLATE. 

% (c) H.J. Lincklaen Arriens, 
%     Delft University of Technology, December 2004
%                                     Februari 2005, latencies differing from 1
%                                                    enabled/corrected.
%									  June 2005, ASL/ASR added


ALUov = 'wrap';

bfSlash = filesep;
fName   = strtok(cirFilename, '.');

opTypesTbl = { 'mul' '*';
			   'add' '+' ;
			   'sub' '-' ;
			   'lt ' '<' ;
			   'le ' '<=';
			   'eq ' '==';
			   'ge ' '>=';
			   'gt ' '>'  };


% read input data
CIRC_Data = parse(cirFilename);
InOuts    = groupIOs(CIRC_Data);

iconnTbl = CIRC_Data.iconnTbl;
[nOps,dummy,shiftInfo] = size(iconnTbl);
if ( shiftInfo ~= 2 )
	iconnTbl(:,:,2) = zeros(nOps,dummy);
end
xconnTbl = CIRC_Data.xconnTbl;
 opTypes = CIRC_Data.opTypes;
 opNames = CIRC_Data.opNames;
inpNames = CIRC_Data.inpNames;
outNames = CIRC_Data.outNames;
dlydInps = CIRC_Data.dlydInps;


schedMethod = lower(schedMethod);
if isempty(varargin)
	delayMUL = 1;
	delayALU = 1;
else
	delayMUL = varargin{1};
	delayALU = varargin{2};
end
if strcmp(schedMethod,'list')
	if ( size(varargin,2) <= 2 )
		nMULs = 1;
		nALUs = 1;
		fprintf( 'LIST: 1 multiplier and 1 ALU assumed ...\n' );
	elseif ( size(varargin,2) == 4 )
		nMULs = varargin{3};
		nALUs = varargin{4};
	else
		error( 'LIST ERROR: Specify number of multipliers AND number of ALU''s ...' );
	end
else
	if ( size(varargin,2) > 2 )
		beep;
		fprintf( '"nMULs" and/or "nALUs" arguments will be ignored (reset to 1) ...\n' );
		fprintf( 'Hit any key to continue or press Ctrl-C to stop ...\n' );
		pause
	end 
	nMULs = 1;
	nALUs = 1;
end


dateStr = datestr(clock);
% for easy passing to function calls, collect in one structure
SSG_Data.filePath   = [ pwd bfSlash cirFilename ];
SSG_Data.inFilename = fName;
SSG_Data.method     = schedMethod;
SSG_Data.delayMUL   = delayMUL;
SSG_Data.delayALU   = delayALU;
SSG_Data.nMULs      = nMULs;
SSG_Data.nALUs      = nALUs;
SSG_Data.dateStr    = dateStr;


assignin( 'base', 'CIRC_Data', CIRC_Data );
assignin( 'base', 'SSG_Data', SSG_Data );


T = schedule(CIRC_Data,SSG_Data);

constCoeffs = InOuts.constCoeffs;
extInps     = InOuts.extInps;
ssgInps     = InOuts.ssgInps;
ssgOutps    = InOuts.ssgOutps;
extOutps    = InOuts.extOutps;
dlydInpsTbl = InOuts.dlydInpsTbl;

nCoeffs   = size(constCoeffs,1); 
nExtInps  = size(extInps,1);     
nSsgInps  = size(ssgInps,1);     
nSsgOutps = size(ssgOutps,1);
nExtOutps = size(extOutps,1);  
nDlydInps = size(dlydInps,1);

if ~exist('toScreen','var') 
	toScreen = 0;
end
if ( toScreen == 0 )
	dsgnDir       = [ pwd bfSlash fName bfSlash ];
	dsgnMATLABDir = [ dsgnDir 'matlab' bfSlash ];
	if exist(dsgnMATLABDir,'dir')
		beep;
		questStr = [ 'Directory ' dsgnMATLABDir ' already exists. Continue and overwrite ?' ];
		respStr  = questdlg(questStr,'Previous Design Found','Yes','Cancel','Cancel');
		if strcmp(respStr,'Cancel')
			error( 'Cancel ...' );
		end
	end

	if ~exist(dsgnDir,'dir')
		mkdir(dsgnDir);
	end
	if ~exist(dsgnMATLABDir,'dir')
		mkdir(dsgnMATLABDir);
	end
else
	fid = 1;
end


% write some info about .INP file and make coeffsSeq available to caller
coeffsSeq = constCoeffs;
if ( nCoeffs ~= 0 )
	fprintf( '\nExpected sequence of coefficients in .INP file:\n' );
	for i = 1:nCoeffs
		fprintf( '%s\n', constCoeffs{i} );
	end
	fprintf( '\n' );
end

% generate a string showing all coefficients
varCoeffsStr = '';
for i = 1:nCoeffs
	varCoeffsStr = [ varCoeffsStr ' ' constCoeffs{i} ];
end
% generate a string showing all delayed (internal) inputs
varDlydInpsStr = '';
for i = 1:nDlydInps
	varDlydInpsStr = [ varDlydInpsStr ' ' dlydInpsTbl{i,1,:} ];
end
% generate a string showing all external inputs
varInpsStr = extInps{1};
for i = 2:nExtInps
	varInpsStr = [ varInpsStr ',' extInps{i} ];
end
% generate a string showing all ssg-inputs
varSsgInpsStr = ssgInps{1};
for i = 2:nSsgInps
	varSsgInpsStr = [ varSsgInpsStr ',' ssgInps{i} ];
end
% generate a string showing all ssg-outputs
varSsgOutpsStr = ssgOutps{1};
for i = 2:nSsgOutps
	varSsgOutpsStr = [ varSsgOutpsStr ',' ssgOutps{i} ];
end
% generate a string showing all external outputs
varOutpsStr = extOutps{1};
for i = 2:nExtOutps
	varOutpsStr = [ varOutpsStr ',' extOutps{i} ];
end


%===============================================================================

% top level function	
if ( toScreen == 0 )
	fid = fopen( [dsgnMATLABDir 'TB_' fName '_auto.m'], 'w' );
end
fprintf( fid, 'function [%s] = TB_%s_auto(SFxd,%s,DBG)\n', ...
													varOutpsStr, fName, varInpsStr );
fprintf( fid, '%%========================================================\n' );
fprintf( fid, '%% next code generated by gen_TB.m, %s\n', dateStr );
commentStr = sprintf( '%s, %d, %d', schedMethod, delayMUL, delayALU );
if strcmp(schedMethod,'list')
	commentStr = sprintf( '%s, %d, %d', commentStr, nMULs, nALUs );
end
fprintf( fid, '%% %s:  %s\n', cirFilename, commentStr );
fprintf( fid, '%%========================================================\n\n' );
	
% first static globals like the coefficients
if ( (nCoeffs ~= 0) || (nDlydInps ~= 0) )
	fprintf( fid, 'global  %s %s\n\n', varCoeffsStr, varDlydInpsStr );
end
% call to local function
fprintf( fid, '[%s] = lfDFG(SFxd,%s,DBG);\n', varSsgOutpsStr, varInpsStr );
% update delayed inputs for this loop
for i = 1:nDlydInps
	fprintf( fid, '%s = %s;\n', dlydInpsTbl{i,1}, dlydInpsTbl{i,2} );
end


%===============================================================================

fprintf( fid, '\n\n%%========================================================\n' );
fprintf( fid, '%%============     Local Functions     ===================\n' );
fprintf( fid, '%%========================================================\n' );
fprintf( fid, 'function [%s] = lfDFG(SFxd,%s,DBG);\n', varSsgOutpsStr, varInpsStr );
if ( (nCoeffs ~= 0) || (nDlydInps ~= 0) )
	fprintf( fid, 'global  %s %s\n\n', varCoeffsStr, varDlydInpsStr );
end

T((T(:,2)==0),2) = T((T(:,2)==0),1);
assignin('base','T',T);
resultStates = unique(T(:,2));

for t = 1:length(resultStates)

	newOps = find( T(:,2) == resultStates(t) );
	fprintf( fid, '  %% ------ after state %d -----\n', resultStates(t) );
	printStr = sprintf( '''  === AFTER STATE %d:\\n''', resultStates(t) );
	fprintf( fid, '  if DBG, fprintf(%s); end\n', printStr );
	for j = 1:length(newOps)
		ixLha  = newOps(j);
		lhaStr = opNames{ixLha};
		for i = 2:3
			ixRha  = iconnTbl(ixLha,i,1);
			if ( ixRha > 0 )
				rhaStr = opNames{ixRha};
			else
				rhaStr = inpNames{-ixRha}; 
			end
			if ( iconnTbl(ixLha,i,2) ~= 0 )
				kBits  = iconnTbl(ixLha,i,2);
				if ( kBits < 0 )
					shiftDir = '<<';
				else
					shiftDir = '>>';
				end
				rhdStr = [ '(' rhaStr shiftDir num2str(abs(kBits)) ')' ];
				rhaStr = [ 'lfShift(' rhaStr ',' num2str(kBits) ',SFxd)' ];
			else
				rhdStr = rhaStr;
			end
			rhiStr(i-1) = {rhaStr};
			rhxStr(i-1) = {rhdStr};
		end
		operStr = lower(opTypes(ixLha,:));
		if strcmp(operStr,'mul')
			dbgStr = sprintf( '%s = %s * %s:', lhaStr, rhxStr{1}, rhxStr{2} );
			fprintf( fid, '  %s = MUL(%s,%s,SFxd, DBG,''%s'');\n', ...
												lhaStr, rhiStr{1}, rhiStr{2}, dbgStr );
		else
			ix = find( strcmp(opTypesTbl(:,1),operStr) );
			operStr = deblank(operStr);		% get rid of spaces at the end of the string
			dbgStr = sprintf( '%s = %s %s %s:', ...
											  lhaStr, rhxStr{1}, opTypesTbl{ix,2}, rhxStr{2} );
			fprintf( fid, '  [%s,overflow] = ALU(%s,%s,SFxd, ''%s'', ''%s'', DBG,''%s'');\n', ...
									lhaStr, rhiStr{1}, rhiStr{2}, operStr, ALUov, dbgStr );
		end

	end
	
end
fprintf( fid, '  %% ------ output assignments -----\n' );
for i = 1:length(outNames)
	ix1 = find(iconnTbl(:,4,1) == i);
	if ~isempty(ix1)
		fromStr = opNames{ix1};
		if ( iconnTbl(ix1,1,2) ~= 0 )
			fromStr = [ 'lfShift(' fromStr ',' num2str(iconnTbl(ix1,1,2)) ',SFxd)' ];
		end
		fprintf( fid, '  %s = %s;\n', outNames{i}, fromStr );
	else
		ix2 = find( strcmp(xconnTbl(:,1),outNames{i}) );
		fprintf( fid, '  %s = %s;\n', outNames{i}, xconnTbl{ix2,2} );
	end		 
end

% next code only if Shift operations have been used
if ( any(any(iconnTbl(:,:,2))) )
	fprintf( fid, '\n%%========================================================\n' );
	fprintf( fid, 'function outVal = lfShift(inVal,kBits,fxd)\n' );
	fprintf( fid, '  %% overflow check and truncation are done in toSFixp\n' );
	fprintf( fid, '  try\n' );
	fprintf( fid, '    outVal = toSFixp(inVal/(2^kBits),fxd);\n' );
	fprintf( fid, '  catch\n' );
	errStr = sprintf( 'SHIFT ERROR: result doesn''''t fit in %%d integer bits ...' );
	fprintf( fid, '    error( ''%s'', N-M );\n', errStr );
	fprintf( fid, '  end\n' );
end

fprintf( fid, '\n%%[EOF]\n\n' );

if ( fid ~= 1 )
	fclose(fid);
end



%==========================================================================================
%=========  ... and ( an example of ) the testbench
%==========================================================================================

if ( toScreen == 0 )
	fid = fopen( [dsgnMATLABDir 'testbench_' fName '_auto.m'], 'w' );
else
	fprintf( '\nTESTBENCH FILE:\n\n' );
end

FName = upper(fName);
fprintf( fid, 'function testbench_%s_auto(SFxd3,DBG)\n', fName );
fprintf( fid, '%%Syntax: TESTBENCH_%s_AUTO(SFxd3,DBG)\n', fName ); 
fprintf( fid, '%%	SFxd3 defines width of signed fractional fixed-point databus\n' ); 
fprintf( fid, '%%	in vector [N M x], where N is the total number of bits of the\n' ); 
fprintf( fid, '%%	external I/O bus from which M are fractional bits.\n' ); 
fprintf( fid, '%%	x defines an additional number of bits to extent N inside the SSG.\n' );  
fprintf( fid, '%%	If DBG = 1, intermediate results are printed.\n' );
fprintf( fid, '%%	Needs ''%s.INP'' in current directory to read input data from,\n', FName ); 
fprintf( fid, '%%	writes ''%s.OUT'' with results.\n\n', FName );

% first static globals like the coefficients
if ( (nCoeffs ~= 0) || (nDlydInps ~= 0) )
	fprintf( fid, 'global  %s %s\n\n', varCoeffsStr, varDlydInpsStr );
end

fprintf( fid, 'N = SFxd3(1);\n' );
fprintf( fid, 'M = SFxd3(2);\n' );
fprintf( fid, 'if ( N == M )\n' );
fprintf( fid, '\terror( ''SFxd leaves no room for a sign bit ...'' );\n' );
fprintf( fid, 'end\n' );
fprintf( fid, 'NX = N + SFxd3(3);\n' );
fprintf( fid, 'fxd  = [N  M];\n' );
fprintf( fid, 'SFxd = [NX M];\n\n' );

% reset delayed inputs
fprintf( fid, '%% reset delayed inputs\n' );
for i = 1:nDlydInps
	fprintf( fid, '%s = 0;\n', dlydInpsTbl{i,1} );
end

fprintf( fid, '\nfidi = fopen( ''%s.INP'', ''r'' );\n', FName );
fprintf( fid, 'if ( fidi == -1 )\n' );
fprintf( fid, '\terror( ''Cannot open file ''''%s.INP'''' ...'' );\n', FName );
fprintf( fid, '\tbeep;\n' );
fprintf( fid, 'end\n' ); 
fprintf( fid, 'fido = fopen( ''%s.OUT'', ''w'' );\n', FName );
str = 'fprintf( fido, ''MATLAB testbench output: %s\n'', datestr(now) );';
fprintf( fid, '%s\n', str );
str = 'fprintf( fido, ''SFxd3 = [%d %d %d]\n'', SFxd3 );';
fprintf( fid, '%s\n', str );
% write first output(s): always zero because of one-sample delay
fprintf( fid, '%% simulate one SAMPLE-clock delay for each output\n' );
str = 'fprintf( fido, ''x"%s"\n'', fixp2hex(0,fxd) );';
for i = 1:nExtOutps
	fprintf( fid, '%s\n', str );
end
fprintf( fid,'\n' );

fprintf( fid, '%% read constants from file\n' );
str = [ 'fprintf( ''Reading file ''''' FName '.INP'''' ...\n'' )' ];
fprintf( fid, '%s;\n', str );
for i = 1:nCoeffs
	fprintf( fid, '%s = lfReadVhdHex(fidi,fxd);\n', constCoeffs{i} );
end
fprintf( fid, '\ni   = 1;\n' );
fprintf( fid, 'EOF = 0;\n' );
fprintf( fid, 'while ~EOF\n' );
fprintf( fid, '\t%% read input value(s) from file\n' );
for i = 1:nExtInps
	fprintf( fid, '\t[%s,EOF] = lfReadVhdHex(fidi,fxd);\n', extInps{i} );
end

fprintf( fid, '\tif ~EOF\n' );
str = 'if DBG,  fprintf(''\nPASS %d:\n'', i);  end';
fprintf( fid, '\t\t%s\n', str );
fprintf( fid, '\t\t%%=======  Call generated circuit function  ============\n' );
fprintf( fid, '\t\t[%s] = TB_%s_auto(SFxd,%s,DBG);\n', varOutpsStr, fName, varInpsStr );
fprintf( fid, '\t\t%% write output value(s) to file and to screen\n' );
for i = 1:nExtOutps
	str = 'fprintf( fido, ''x"%s"\n''';
	fprintf( fid, '\t\t%s, fixp2hex(%s,fxd) );\n', str, extOutps{i} );
	str = [ 'fprintf( ''' extOutps{i} ':  x"%s"   %#- 10g\n''' ];
	fprintf( fid, '\t\t%s, fixp2hex(%s,fxd), %s );\n', str, extOutps{i}, extOutps{i} ); 
end

fprintf( fid, '\t\ti = i +1;\n' );
fprintf( fid, '\tend\n' );
fprintf( fid, 'end\n' );
fprintf( fid, 'fclose(''all'');\n' );
str = [ 'fprintf( ''written file ''''' FName '.OUT''''\n\n'' )' ];
fprintf( fid, '%s;\n\n\n', str );


fprintf( fid, '%%================================================================================\n' );
fprintf( fid, '%%=======   LOCAL FUNCTIONS   ==================================================== \n' );
fprintf( fid, '%%================================================================================\n' );
fprintf( fid, 'function [outInt,EOF] = lfReadVhdHex(fid,fxd)\n' );
fprintf( fid, '  inLine = fgetl(fid);\n' );
fprintf( fid, '  %% skip blank lines\n' );
fprintf( fid, '  while ( ischar(inLine) && (length(inLine) < 4) )\n' );
fprintf( fid, '    inLine = fgetl(fid);\n' );
fprintf( fid, '  end\n' );
fprintf( fid, '  %% skip if not hex input line\n' );
fprintf( fid, '  while ( ischar(inLine) && (inLine(1) ~= ''x'') )\n' );
fprintf( fid, '    inLine = fgetl(fid);\n' );
fprintf( fid, '  end\n' );
fprintf( fid, '  if ischar(inLine)\n' );
fprintf( fid, '    dQuotePos = find( inLine == ''"'' );\n' );
fprintf( fid, '    if ( length(dQuotePos) ~= 2 )\n' );
str = 'INP FILE: %s, incorrect hex input format ...';
fprintf( fid, '      error( ''%s'', inLine );\n', str );
fprintf( fid, '    end\n' );
fprintf( fid, '    ix     = dQuotePos + [ 1 -1 ];\n' );
fprintf( fid, '    outInt = hex2fixp( inLine(ix(1):ix(2)),fxd );\n' );
fprintf( fid, '    EOF    = 0;\n' );
fprintf( fid, '  else\n' );
fprintf( fid, '    outInt = -1;\n' );
fprintf( fid, '    EOF    =  1;\n' );
fprintf( fid, '  end\n\n%%[EOF]\n\n' );


if ( fid ~= 1 )
	fclose(fid);
end


    