function tFrames = listSched(adjMat,delayVec,mulOps,nMULs,nALUs)
%LISTSCHED	Find the SSG using a List Scheduling method
%	tFrames = LISTSCHED(adjMat,delayVec,mulOps,nMULs,nALUs)
%	returns a two-column vector 'tFrames' describing the 
%	SSG (Scheduled Sequencing Graph) of the operations when
%	scheduled according to a specified List Scheduling method under
%	resource constraints.
%	The unscheduled flow has to be given in	'adjMat' with the latency 
%	per resource type (e.g. MUL, ALU) as integer number of clock cycles 
%	specified in 'delayVec'. 
%	In 'mulOps' the resource type of the operations should be given,
%	and the available numbers of resources in nMULs and nALUs.
%
%	See also ASAP, ALAP, FORCED, TOADJMAT, PARSE.  

% (c) H.J. Lincklaen Arriens, 
%     Delft University of Technology, November 2004
%
%	  Apr 2005, problem with number of MULs|ALUs exceeding maximum 
%				number in case of delays > 1 has been solved, Huib


T        = asap(adjMat,delayVec);
minTs    = T(:,1);
% compute a priority value for each node, depending on the
% length of the vertices between the node itself and the sink ...
% N.B.! equals time frame of ALAP
T        = alap(adjMat,delayVec);
priorVec = T(:,1);

nNodes   = length(minTs);

% perform scheduling 
delayMat      = repmat(delayVec,1,nNodes) .* adjMat;
tFrames       = zeros(nNodes,2);
pendingMULs   = zeros(nNodes,1);
pendingALUs   = zeros(nNodes,1);

scheduledMULs = lfSchedule( delayMat,priorVec, mulOps,nMULs,sum(pendingMULs) );
scheduledALUs = lfSchedule( delayMat,priorVec,~mulOps,nALUs,sum(pendingALUs) );
scheduleNow   = [ scheduledMULs; scheduledALUs ];
pendingNodes  = any(scheduleNow);

t = 1;
while pendingNodes

	tFrames(scheduleNow,1)         = t;
	tFrames( (pendingMULs > 0), 2) = t;
	tFrames( (pendingALUs > 0), 2) = t;
	pendingMULs(scheduledMULs) = delayVec(scheduledMULs);
	pendingALUs(scheduledALUs) = delayVec(scheduledALUs);
	delayMat(:,scheduleNow)    = NaN;
	% adjust for time step
	ixMULs              = find(pendingMULs > 0);
	pendingMULs(ixMULs) = pendingMULs(ixMULs)   -1;
	ixALUs              = find(pendingALUs > 0);
	pendingALUs(ixALUs) = pendingALUs(ixALUs)   -1;
	pendingOps          = [ ixMULs; ixALUs ];
	delayMat(pendingOps,:)   = delayMat(pendingOps,:) -1;
	delayMat( delayMat < 0 ) = 0;
	% next time step
	t = t + 1;
	scheduledMULs = lfSchedule( delayMat,priorVec, mulOps,nMULs,sum(pendingMULs) );
	scheduledALUs = lfSchedule( delayMat,priorVec,~mulOps,nALUs,sum(pendingALUs) );
	scheduleNow   = [ scheduledMULs; scheduledALUs ];
	pendingNodes  = ( any(scheduleNow) | any(pendingOps) );

end


%================================================================================
%=======   LOCAL FUNCTIONS   ==================================================== 
%================================================================================
function scheduledOps = lfSchedule(delayMat,priorVec,theseOps,nOps,nPendingOps)
	if ( (nOps - nPendingOps) > 0 )
		scheduleQueue = ( sum(delayMat) == 0 )';
		candidateOps  = find( (theseOps & scheduleQueue) == 1 );
		[dummy,ix]    = sort( priorVec(candidateOps) );
		candidateOps  = candidateOps(ix);
		scheduledOps  = candidateOps( 1 : min(nOps-nPendingOps,length(candidateOps)) );
	else
		scheduledOps = [];
	end
