function [tFrames,distrib] = schedule(CIRC_Data, SSG_Data)
%SCHEDULE	For internal use only
%	tFrames = SCHEDULE(CIRC_Data, SSG_Data) determines a time schedule given
%	the parameters in the structures CIRC_Data and SSG_Data:
%		CIRC_Data.iconnTbl = iconnTbl;
%		CIRC_Data.xconnTbl = xconnTbl;  <-- not used here
%		CIRC_Data.opTypes  = opTypes;
%		CIRC_Data.opNames  = opNames; 
%		CIRC_Data.inpNames = inpNames;  <-- not used here
%		CIRC_Data.outNames = outNames;  <-- not used here
%		CIRC_Data.dlydInps = dlydInps;  <-- not used here
%
%		SSG_Data.filePath   = fPath;    <-- not used here
%		SSG_Data.inFilename = fName;    
%		SSG_Data.method     = schedMethod;
%		SSG_Data.delayMUL   = delayMUL;
%		SSG_Data.delayALU   = delayALU;
%		SSG_Data.nMULs      = nMULs;
%		SSG_Data.nALUs      = nALUs;
%		SSG_Data.dateStr    = datestr(clock);
%
%	'schedMethod' can be 'ASAP','ALAP','FORCED' or 'LIST', using
%	two resources, e.g. MUL and ALU (add,sub,cmparisons).
%	Time delays are multiple integers of one clock cycle.
%	A file 'graph.dot' will be created in the current directory.
%
%	[tFrames,distrib] = SCHEDULE(CIRC_Data, SSG_Data)
%	'distrib' indicates the distribution of the MULs and ALUs over the time states.
%
%	See also  PARSE, GROUPIOS.

% (c) H.J. Lincklaen Arriens, 
%     Delft University of Technology, November 2004
%									  June 2005, << and >> added


% copy variables to be used here from CIRC_Data
iconnTbl    = CIRC_Data.iconnTbl;
opTypes     = CIRC_Data.opTypes;
opNames     = CIRC_Data.opNames; 
% ... and from SSG_Data
fName       = SSG_Data.inFilename;
schedMethod = SSG_Data.method;    
delayMUL    = SSG_Data.delayMUL;  
delayALU    = SSG_Data.delayALU;  
nMULs       = SSG_Data.nMULs;     
nALUs       = SSG_Data.nALUs;     
dateStr     = SSG_Data.dateStr;


adjMat = toAdjMat(iconnTbl);

[nNodes,dummy,shiftInfo] = size(iconnTbl);
if ( shiftInfo ~= 2 )
	iconnTbl(:,:,2) = zeros(nNodes,dummy);
end
if ( size(opTypes,1) ~= nNodes )
	error( 'input data mismatch ...' );
end
mulOps  = zeros(nNodes,1);
opTypes = lower(opTypes);
for i = 1:nNodes
	mulOps(i) = strcmp(opTypes(i,:),'mul');
end

delayVec = delayALU(ones(nNodes,1));
delayVec(mulOps == 1) = delayMUL;

assignin( 'base', 'CIRC_Data', CIRC_Data );
assignin( 'base', 'SSG_Data', SSG_Data );
 
switch lower(schedMethod)
	case 'asap', 	T = asap(adjMat,delayVec);
	case 'alap', 	T = alap(adjMat,delayVec);
	case 'list',    T = listSched(adjMat,delayVec,mulOps,nMULs,nALUs);
	case 'forced',  if ( (delayMUL==1) && (delayALU==1) )
						T = forceD(adjMat,mulOps,opNames);
					else
						error( 'ForceD not yet implemented for arbitrary delays ...' );
					end
	otherwise,		error('Unknown method ...' );
end

tFrames = T;


%====================================================================================
% Write graph.dot 
%====================================================================================

fid = fopen('graph.dot','w');

fprintf( fid, '/* %s, generated by ''schedule.m'' */\n', dateStr );
commentStr = sprintf( '%s, %d, %d', schedMethod, delayMUL, delayALU );
if strcmp(schedMethod,'list')
	commentStr = sprintf( '%s, %d, %d', commentStr, nMULs, nALUs );
end
fprintf( fid, '/* %s:  %s */\n\n', fName, commentStr );

fprintf( fid, 'digraph G {\n' );	% 'G' needed buy 'dotty'
%fprintf( fid, 'digraph G {\n\tranksep = 0.75;\n' );
fprintf( fid, '\t{\n\t\tnode [shape=plaintext,fontname=Courier,fontsize=11];\n' );
tmpStr = '\t\t';
nFrames = max(max(T));
for i = 1:nFrames-1
	tmpStr = [ tmpStr sprintf( '"STATE %d" -> ', i ) ];
end
fprintf( fid, [ tmpStr '"STATE %d";\n\t}\n\n' ], nFrames );
	
tmpStr = 'shape=circle,height=0.6,fixedsize=true,fontname=Courier,fontsize=10';
fprintf( fid, '\t{\n\t\tnode [%s]\n', tmpStr );
for i = 1:nNodes
	if ( delayVec(i) == 1 )		% unit delay 
		fprintf( fid, '\t\tn%d [label="%s\\n%s"];\n', i, char(opNames(i)), opTypes(i,:) );
	else
		for j = 1:delayVec(i)
			s = char('a'+j-1);
			fprintf( fid, '\t\tn%d%s [label="%s%s\\n%s",color=red];\n', ...
													  	i, s, char(opNames(i)), s, opTypes(i,:) );
		end
	end	
end
fprintf( fid, '\t}\n\n' );

rankStr  = cell(nFrames,1);
tDistrib = cell(nFrames,1);

for i = 1:nNodes
	ix1 = T(i,1);
	if ( delayVec(i) == 1 )
		rankStr{ix1}  = [ rankStr{ix1}  sprintf( '"n%d"; ', i ) ];
		tDistrib{ix1} = [ tDistrib{ix1} i ];
	else
		for	j = T(i,1):T(i,2)
			rankStr{j}  = [ rankStr{j} sprintf( '"n%d%s"; ', i, char('a'+j-ix1) ) ];
			tDistrib{j} = [ tDistrib{j} i ];
		end
	end
end
for i = 1:nFrames
	fprintf( fid, [ '\t{ rank= same; "STATE %d"; ' rankStr{i} '}\n' ], i );
end
fprintf( fid, '\n' );

% compute precedence
[r,c]  = find( adjMat == 1 );
[r,ix] = sort(r);
v = [ r c(ix) ];

for i=1:nNodes
	if ( delayVec(i) ~= 1 )
		for j = 0:delayVec(i)-2
			fprintf( fid, '\tn%d%s -> n%d%s [color=red];\n', ...
													i, char('a'+j), i, char('b'+j) );
		end
	end
end
for i = 1:size(v,1)
	fromNode = v(i,1);
	toNode   = v(i,2);
	fromStr  = sprintf( 'n%d', fromNode );
	toStr    = sprintf( 'n%d', toNode   );
	if ( delayVec(fromNode) ~= 1 )
		fromStr = sprintf( '%s%s', fromStr, char('a'+T(fromNode,2)-T(fromNode,1)) );
	end
	if ( delayVec(toNode) ~= 1 )
		toStr = sprintf( '%sa', toStr );
	end
	inNo = find( iconnTbl(toNode,2:3,1) == fromNode );
	if ( iconnTbl(toNode,inNo+1,2) == 0 )	% no shift
		fprintf( fid, '\t%s -> %s;\n', fromStr, toStr );
	else
		kBits = iconnTbl(toNode,inNo+1,2);
		if ( kBits < 0 )
			shiftStr = sprintf( '<<%d', -kBits );
		else 
			shiftStr = sprintf( '>>%d',  kBits );
		end
		fprintf( fid, '\t%s -> %s [style=bold,label="%s"];\n', fromStr, toStr, shiftStr );
	end		
end
fprintf( fid, '\n' );	
fprintf( fid, '}\n' );
fclose( fid );

distrib = zeros(nFrames,2);
for i = 1:nFrames
	testVec = zeros(nNodes,1);
	testVec(tDistrib{i}) = 1;
	distrib(i,1) = sum( and(testVec, mulOps) );
	distrib(i,2) = sum( and(testVec,~mulOps) );
end


